/*
* Copyright (c) 2006-2009 Erin Catto http://www.box2d.org
*
* This software is provided 'as-is', without any express or implied
* warranty.  In no event will the authors be held liable for any damages
* arising from the use of this software.
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
* 1. The origin of this software must not be misrepresented; you must not
* claim that you wrote the original software. If you use this software
* in a product, an acknowledgment in the product documentation would be
* appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
* misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution.
*/

goog.provide('box2d.b2Contact');

goog.require('box2d.b2Settings');
goog.require('box2d.b2Collision');
goog.require('box2d.b2TimeOfImpact');

/**
 * Friction mixing law. The idea is to allow either fixture to
 * drive the restitution to zero. For example, anything slides
 * on ice.
 * @export
 * @return {number}
 * @param {number} friction1
 * @param {number} friction2
 */
box2d.b2MixFriction = function (friction1, friction2)
{
	return box2d.b2Sqrt(friction1 * friction2);
}

/**
 * Restitution mixing law. The idea is allow for anything to
 * bounce off an inelastic surface. For example, a superball
 * bounces on anything.
 * @export
 * @return {number}
 * @param {number} restitution1
 * @param {number} restitution2
 */
box2d.b2MixRestitution = function (restitution1, restitution2)
{
	return restitution1 > restitution2 ? restitution1 : restitution2;
}

/**
 * A contact edge is used to connect bodies and contacts
 * together in a contact graph where each body is a node and
 * each contact is an edge. A contact edge belongs to a doubly
 * linked list maintained in each attached body. Each contact
 * has two contact nodes, one for each attached body.
 * @export
 * @constructor
 */
box2d.b2ContactEdge = function ()
{
};

/**
 * @export
 * @type {box2d.b2Body}
 */
box2d.b2ContactEdge.prototype.other = null; ///< provides quick access to the other body attached.
/**
 * @export
 * @type {box2d.b2Contact}
 */
box2d.b2ContactEdge.prototype.contact = null; ///< the contact
/**
 * @export
 * @type {box2d.b2ContactEdge}
 */
box2d.b2ContactEdge.prototype.prev = null; ///< the previous contact edge in the body's contact list
/**
 * @export
 * @type {box2d.b2ContactEdge}
 */
box2d.b2ContactEdge.prototype.next = null; ///< the next contact edge in the body's contact list

/**
 * Flags stored in m_flags
 * @enum
 */
box2d.b2ContactFlag =
{
	e_none			: 0,
	e_islandFlag	: 0x0001, /// Used when crawling contact graph when forming islands.
	e_touchingFlag	: 0x0002, /// Set when the shapes are touching.
	e_enabledFlag	: 0x0004, /// This contact can be disabled (by user)
	e_filterFlag	: 0x0008, /// This contact needs filtering because a fixture filter was changed.
	e_bulletHitFlag	: 0x0010, /// This bullet contact had a TOI event
	e_toiFlag		: 0x0020  /// This contact has a valid TOI in m_toi
};
goog.exportProperty(box2d.b2ContactFlag, 'e_none'         , box2d.b2ContactFlag.e_none         );
goog.exportProperty(box2d.b2ContactFlag, 'e_islandFlag'   , box2d.b2ContactFlag.e_islandFlag   );
goog.exportProperty(box2d.b2ContactFlag, 'e_touchingFlag' , box2d.b2ContactFlag.e_touchingFlag );
goog.exportProperty(box2d.b2ContactFlag, 'e_enabledFlag'  , box2d.b2ContactFlag.e_enabledFlag  );
goog.exportProperty(box2d.b2ContactFlag, 'e_filterFlag'   , box2d.b2ContactFlag.e_filterFlag   );
goog.exportProperty(box2d.b2ContactFlag, 'e_bulletHitFlag', box2d.b2ContactFlag.e_bulletHitFlag);
goog.exportProperty(box2d.b2ContactFlag, 'e_toiFlag'      , box2d.b2ContactFlag.e_toiFlag      );

/**
 * The class manages contact between two shapes. A contact
 * exists for each overlapping AABB in the broad-phase (except
 * if filtered). Therefore a contact object may exist that has
 * no contact points.
 * @export
 * @constructor
 */
box2d.b2Contact = function ()
{
	this.m_nodeA = new box2d.b2ContactEdge();
	this.m_nodeB = new box2d.b2ContactEdge();
	this.m_manifold = new box2d.b2Manifold();
	this.m_oldManifold = new box2d.b2Manifold();
}

/**
 * @export
 * @type {box2d.b2ContactFlag}
 */
box2d.b2Contact.prototype.m_flags = box2d.b2ContactFlag.e_none;

/**
 * World pool and list pointers.
 * @export
 * @type {box2d.b2Contact}
 */
box2d.b2Contact.prototype.m_prev = null;
/**
 * @export
 * @type {box2d.b2Contact}
 */
box2d.b2Contact.prototype.m_next = null;

/**
 * Nodes for connecting bodies.
 * @export
 * @type {box2d.b2ContactEdge}
 */
box2d.b2Contact.prototype.m_nodeA = null;
/**
 * @export
 * @type {box2d.b2ContactEdge}
 */
box2d.b2Contact.prototype.m_nodeB = null;

/**
 * @export
 * @type {box2d.b2Fixture}
 */
box2d.b2Contact.prototype.m_fixtureA = null;
/**
 * @export
 * @type {box2d.b2Fixture}
 */
box2d.b2Contact.prototype.m_fixtureB = null;

/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_indexA = 0;
/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_indexB = 0;

/**
 * @export
 * @type {box2d.b2Manifold}
 */
box2d.b2Contact.prototype.m_manifold = null;

/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_toiCount = 0;
/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_toi = 0;

/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_friction = 0;
/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_restitution = 0;

/**
 * @export
 * @type {number}
 */
box2d.b2Contact.prototype.m_tangentSpeed = 0;

/**
 * @export
 * @type {box2d.b2Manifold}
 */
box2d.b2Contact.prototype.m_oldManifold = null;

/**
 * Get the contact manifold. Do not modify the manifold unless
 * you understand the internals of Box2D.
 * @export
 * @return {box2d.b2Manifold}
 */
box2d.b2Contact.prototype.GetManifold = function ()
{
	return this.m_manifold;
}

/**
 * Get the world manifold.
 * @export
 * @return {void}
 * @param {box2d.b2WorldManifold} worldManifold
 */
box2d.b2Contact.prototype.GetWorldManifold = function (worldManifold)
{
	var bodyA = this.m_fixtureA.GetBody();
	var bodyB = this.m_fixtureB.GetBody();
	var shapeA = this.m_fixtureA.GetShape();
	var shapeB = this.m_fixtureB.GetShape();
	worldManifold.Initialize(this.m_manifold, bodyA.GetTransform(), shapeA.m_radius, bodyB.GetTransform(), shapeB.m_radius);
}

/**
 * Is this contact touching?
 * @export
 * @return {boolean}
 */
box2d.b2Contact.prototype.IsTouching = function ()
{
	return (this.m_flags & box2d.b2ContactFlag.e_touchingFlag) == box2d.b2ContactFlag.e_touchingFlag;
}

/**
 * Enable/disable this contact. This can be used inside the
 * pre-solve contact listener. The contact is only disabled for
 * the current time step (or sub-step in continuous collisions).
 * @export
 * @return {void}
 * @param {boolean} flag
 */
box2d.b2Contact.prototype.SetEnabled = function (flag)
{
	if (flag)
	{
		this.m_flags |= box2d.b2ContactFlag.e_enabledFlag;
	}
	else
	{
		this.m_flags &= ~box2d.b2ContactFlag.e_enabledFlag;
	}
}

/**
 * Has this contact been disabled?
 * @export
 * @return {boolean}
 */
box2d.b2Contact.prototype.IsEnabled = function ()
{
	return (this.m_flags & box2d.b2ContactFlag.e_enabledFlag) == box2d.b2ContactFlag.e_enabledFlag;
}

/**
 * Get the next contact in the world's contact list.
 * @export
 * @return {box2d.b2Contact}
 */
box2d.b2Contact.prototype.GetNext = function ()
{
	return this.m_next;
}

/**
 * Get fixture A in this contact.
 * @export
 * @return {box2d.b2Fixture}
 */
box2d.b2Contact.prototype.GetFixtureA = function ()
{
	return this.m_fixtureA;
}

/**
 * @export
 * @return {number}
 */
box2d.b2Contact.prototype.GetChildIndexA = function ()
{
	return this.m_indexA;
}

/**
 * Get fixture B in this contact.
 * @export
 * @return {box2d.b2Fixture}
 */
box2d.b2Contact.prototype.GetFixtureB = function ()
{
	return this.m_fixtureB;
}

/**
 * @export
 * @return {number}
 */
box2d.b2Contact.prototype.GetChildIndexB = function ()
{
	return this.m_indexB;
}

/**
 * Evaluate this contact with your own manifold and transforms.
 * @export
 * @return {void}
 * @param {box2d.b2Manifold} manifold
 * @param {box2d.b2Transform} xfA
 * @param {box2d.b2Transform} xfB
 */
box2d.b2Contact.prototype.Evaluate = function (manifold, xfA, xfB)
{
}

/**
 * Flag this contact for filtering. Filtering will occur the
 * next time step.
 * @export
 * @return {void}
 */
box2d.b2Contact.prototype.FlagForFiltering = function ()
{
	this.m_flags |= box2d.b2ContactFlag.e_filterFlag;
}

/**
 * Override the default friction mixture. You can call this in
 * box2d.b2ContactListener::PreSolve.
 * This value persists until set or reset.
 * @export
 * @return {void}
 * @param {number} friction
 */
box2d.b2Contact.prototype.SetFriction = function (friction)
{
	this.m_friction = friction;
}

/**
 * Get the friction.
 * @export
 * @return {number}
 */
box2d.b2Contact.prototype.GetFriction = function ()
{
	return this.m_friction;
}

/**
 * Reset the friction mixture to the default value.
 * @export
 * @return {void}
 */
box2d.b2Contact.prototype.ResetFriction = function ()
{
	this.m_friction = box2d.b2MixFriction(this.m_fixtureA.m_friction, this.m_fixtureB.m_friction);
}

/**
 * Override the default restitution mixture. You can call this
 * in box2d.b2ContactListener::PreSolve.
 * The value persists until you set or reset.
 * @export
 * @return {void}
 * @param {number} restitution
 */
box2d.b2Contact.prototype.SetRestitution = function (restitution)
{
	this.m_restitution = restitution;
}

/**
 * Get the restitution.
 * @export
 * @return {number}
 */
box2d.b2Contact.prototype.GetRestitution = function ()
{
	return this.m_restitution;
}

/**
 * Reset the restitution to the default value.
 * @export
 * @return {void}
 */
box2d.b2Contact.prototype.ResetRestitution = function ()
{
	this.m_restitution = box2d.b2MixRestitution(this.m_fixtureA.m_restitution, this.m_fixtureB.m_restitution);
}

/**
 * Set the desired tangent speed for a conveyor belt behavior.
 * In meters per second.
 * @export
 * @return {void}
 * @param {number} speed
 */
box2d.b2Contact.prototype.SetTangentSpeed = function (speed)
{
	this.m_tangentSpeed = speed;
}

/**
 * Get the desired tangent speed. In meters per second.
 * @export
 * @return {number}
 */
box2d.b2Contact.prototype.GetTangentSpeed = function ()
{
	return this.m_tangentSpeed;
}

/**
 * @export
 * @return {void}
 * @param {box2d.b2Fixture} fixtureA
 * @param {box2d.b2Fixture} fixtureB
 */
box2d.b2Contact.prototype.Reset = function (fixtureA, indexA, fixtureB, indexB)
{
	this.m_flags = box2d.b2ContactFlag.e_enabledFlag;

	this.m_fixtureA = fixtureA;
	this.m_fixtureB = fixtureB;

	this.m_indexA = indexA;
	this.m_indexB = indexB;

	this.m_manifold.pointCount = 0;

	this.m_prev = null;
	this.m_next = null;

	this.m_nodeA.contact = null;
	this.m_nodeA.prev = null;
	this.m_nodeA.next = null;
	this.m_nodeA.other = null;

	this.m_nodeB.contact = null;
	this.m_nodeB.prev = null;
	this.m_nodeB.next = null;
	this.m_nodeB.other = null;

	this.m_toiCount = 0;

	this.m_friction = box2d.b2MixFriction(this.m_fixtureA.m_friction, this.m_fixtureB.m_friction);
	this.m_restitution = box2d.b2MixRestitution(this.m_fixtureA.m_restitution, this.m_fixtureB.m_restitution);
}

/**
 * Update the contact manifold and touching status.
 * Note: do not assume the fixture AABBs are overlapping or are
 * valid.
 * @export
 * @return {void}
 * @param {box2d.b2ContactListener} listener
 */
box2d.b2Contact.prototype.Update = function (listener)
{
	var tManifold = this.m_oldManifold;
	this.m_oldManifold = this.m_manifold;
	this.m_manifold = tManifold;

	// Re-enable this contact.
	this.m_flags |= box2d.b2ContactFlag.e_enabledFlag;

	var touching = false;
	var wasTouching = (this.m_flags & box2d.b2ContactFlag.e_touchingFlag) == box2d.b2ContactFlag.e_touchingFlag;

	var sensorA = this.m_fixtureA.IsSensor();
	var sensorB = this.m_fixtureB.IsSensor();
	var sensor = sensorA || sensorB;

	var bodyA = this.m_fixtureA.GetBody();
	var bodyB = this.m_fixtureB.GetBody();
	var xfA = bodyA.GetTransform();
	var xfB = bodyB.GetTransform();

//	var aabbOverlap = box2d.b2TestOverlapAABB(this.m_fixtureA.GetAABB(0), this.m_fixtureB.GetAABB(0));

	// Is this contact a sensor?
	if (sensor)
	{
//		if (aabbOverlap)
//		{
			var shapeA = this.m_fixtureA.GetShape();
			var shapeB = this.m_fixtureB.GetShape();
			touching = box2d.b2TestOverlapShape(shapeA, this.m_indexA, shapeB, this.m_indexB, xfA, xfB);
//		}

		// Sensors don't generate manifolds.
		this.m_manifold.pointCount = 0;
	}
	else
	{
//		if (aabbOverlap)
//		{
			this.Evaluate(this.m_manifold, xfA, xfB);
			touching = this.m_manifold.pointCount > 0;

			// Match old contact ids to new contact ids and copy the
			// stored impulses to warm start the solver.
			for (var i = 0; i < this.m_manifold.pointCount; ++i)
			{
				var mp2 = this.m_manifold.points[i];
				mp2.normalImpulse = 0;
				mp2.tangentImpulse = 0;
				var id2 = mp2.id;

				for (var j = 0; j < this.m_oldManifold.pointCount; ++j)
				{
					var mp1 = this.m_oldManifold.points[j];

					if (mp1.id.key == id2.key)
					{
						mp2.normalImpulse = mp1.normalImpulse;
						mp2.tangentImpulse = mp1.tangentImpulse;
						break;
					}
				}
			}
//		}
//		else
//		{
//			this.m_manifold.pointCount = 0;
//		}

		if (touching != wasTouching)
		{
			bodyA.SetAwake(true);
			bodyB.SetAwake(true);
		}
	}

	if (touching)
	{
		this.m_flags |= box2d.b2ContactFlag.e_touchingFlag;
	}
	else
	{
		this.m_flags &= ~box2d.b2ContactFlag.e_touchingFlag;
	}

	if (wasTouching == false && touching == true && listener)
	{
		listener.BeginContact(this);
	}

	if (wasTouching == true && touching == false && listener)
	{
		listener.EndContact(this);
	}

	if (sensor == false && touching && listener)
	{
		listener.PreSolve(this, this.m_oldManifold);
	}
}

/**
 * @export
 * @return {number}
 * @param {box2d.b2Sweep} sweepA
 * @param {box2d.b2Sweep} sweepB
 */
box2d.b2Contact.prototype.ComputeTOI = function (sweepA, sweepB)
{
	var input = box2d.b2Contact.prototype.ComputeTOI.s_input;
	input.proxyA.SetShape(this.m_fixtureA.GetShape(), this.m_indexA);
	input.proxyB.SetShape(this.m_fixtureB.GetShape(), this.m_indexB);
	input.sweepA.Copy(sweepA);
	input.sweepB.Copy(sweepB);
	input.tMax = box2d.b2_linearSlop;

	var output = box2d.b2Contact.prototype.ComputeTOI.s_output;

	box2d.b2TimeOfImpact(output, input);

	return output.t;
}
box2d.b2Contact.prototype.ComputeTOI.s_input = new box2d.b2TOIInput();
box2d.b2Contact.prototype.ComputeTOI.s_output = new box2d.b2TOIOutput();

/**
 * @export
 * @constructor
 */
box2d.b2ContactRegister = function ()
{
};

box2d.b2ContactRegister.prototype.createFcn = null;
box2d.b2ContactRegister.prototype.destroyFcn = null;
box2d.b2ContactRegister.prototype.primary = false;

/**
 * @export
 * @constructor
 * @param allocator
 */
box2d.b2ContactFactory = function (allocator)
{
	this.m_allocator = allocator;
	this.InitializeRegisters();
}

box2d.b2ContactFactory.prototype.m_allocator = null;

/**
 * @export
 * @return {void}
 * @param createFcn
 * @param destroyFcn
 * @param {box2d.b2ShapeType} type1
 * @param {box2d.b2ShapeType} type2
 */
box2d.b2ContactFactory.prototype.AddType = function (createFcn, destroyFcn, type1, type2)
{
	if (true)
	{
		var pool = box2d.b2MakeArray(256, function (i) { return createFcn(); } ); // TODO: b2Settings

		var poolCreateFcn = function (allocator)
		{
			if (pool.length > 0)
			{
				return pool.pop();
			}

			return createFcn(allocator);
		}

		var poolDestroyFcn = function (contact, allocator)
		{
			pool.push(contact);
		}

		this.m_registers[type1][type2].pool = pool;
		this.m_registers[type1][type2].createFcn = poolCreateFcn;
		this.m_registers[type1][type2].destroyFcn = poolDestroyFcn;
		this.m_registers[type1][type2].primary = true;

		if (type1 != type2)
		{
			this.m_registers[type2][type1].pool = pool;
			this.m_registers[type2][type1].createFcn = poolCreateFcn;
			this.m_registers[type2][type1].destroyFcn = poolDestroyFcn;
			this.m_registers[type2][type1].primary = false;
		}
	}

	/*
	this.m_registers[type1][type2].createFcn = createFcn;
	this.m_registers[type1][type2].destroyFcn = destroyFcn;
	this.m_registers[type1][type2].primary = true;

	if (type1 != type2)
	{
		this.m_registers[type2][type1].createFcn = createFcn;
		this.m_registers[type2][type1].destroyFcn = destroyFcn;
		this.m_registers[type2][type1].primary = false;
	}
	*/
}

/**
 * @export
 * @return {void}
 */
box2d.b2ContactFactory.prototype.InitializeRegisters = function ()
{
	this.m_registers = Array(box2d.b2ShapeType.e_shapeTypeCount);

	for (var i = 0; i < box2d.b2ShapeType.e_shapeTypeCount; i++)
	{
		this.m_registers[i] = Array(box2d.b2ShapeType.e_shapeTypeCount);

		for (var j = 0; j < box2d.b2ShapeType.e_shapeTypeCount; j++)
		{
			this.m_registers[i][j] = new box2d.b2ContactRegister();
		}
	}

	this.AddType(box2d.b2CircleContact.Create, box2d.b2CircleContact.Destroy, box2d.b2ShapeType.e_circleShape, box2d.b2ShapeType.e_circleShape);
	this.AddType(box2d.b2PolygonAndCircleContact.Create, box2d.b2PolygonAndCircleContact.Destroy, box2d.b2ShapeType.e_polygonShape, box2d.b2ShapeType.e_circleShape);
	this.AddType(box2d.b2PolygonContact.Create, box2d.b2PolygonContact.Destroy, box2d.b2ShapeType.e_polygonShape, box2d.b2ShapeType.e_polygonShape);
	this.AddType(box2d.b2EdgeAndCircleContact.Create, box2d.b2EdgeAndCircleContact.Destroy, box2d.b2ShapeType.e_edgeShape, box2d.b2ShapeType.e_circleShape);
	this.AddType(box2d.b2EdgeAndPolygonContact.Create, box2d.b2EdgeAndPolygonContact.Destroy, box2d.b2ShapeType.e_edgeShape, box2d.b2ShapeType.e_polygonShape);
	this.AddType(box2d.b2ChainAndCircleContact.Create, box2d.b2ChainAndCircleContact.Destroy, box2d.b2ShapeType.e_chainShape, box2d.b2ShapeType.e_circleShape);
	this.AddType(box2d.b2ChainAndPolygonContact.Create, box2d.b2ChainAndPolygonContact.Destroy, box2d.b2ShapeType.e_chainShape, box2d.b2ShapeType.e_polygonShape);
}

/**
 * @export
 * @return {box2d.b2Contact}
 * @param {box2d.b2Fixture} fixtureA
 * @param {number} indexA
 * @param {box2d.b2Fixture} fixtureB
 * @param {number} indexB
 */
box2d.b2ContactFactory.prototype.Create = function (fixtureA, indexA, fixtureB, indexB)
{
	var type1 = fixtureA.GetType();
	var type2 = fixtureB.GetType();

	if (box2d.ENABLE_ASSERTS) { box2d.b2Assert(0 <= type1 && type1 < box2d.b2ShapeType.e_shapeTypeCount); }
	if (box2d.ENABLE_ASSERTS) { box2d.b2Assert(0 <= type2 && type2 < box2d.b2ShapeType.e_shapeTypeCount); }

	var reg = this.m_registers[type1][type2];

	var createFcn = reg.createFcn;
	if (createFcn != null)
	{
		if (reg.primary)
		{
			var c = createFcn(this.m_allocator);
			c.Reset(fixtureA, indexA, fixtureB, indexB);
			return c;
		}
		else
		{
			var c = createFcn(this.m_allocator);
			c.Reset(fixtureB, indexB, fixtureA, indexA);
			return c;
		}
	}
	else
	{
		return null;
	}
}

/**
 * @export
 * @return {void}
 * @param {box2d.b2Contact} contact
 */
box2d.b2ContactFactory.prototype.Destroy = function (contact)
{
	var fixtureA = contact.m_fixtureA;
	var fixtureB = contact.m_fixtureB;

	if (contact.m_manifold.pointCount > 0 &&
		fixtureA.IsSensor() == false &&
		fixtureB.IsSensor() == false)
	{
		fixtureA.GetBody().SetAwake(true);
		fixtureB.GetBody().SetAwake(true);
	}

	var typeA = fixtureA.GetType();
	var typeB = fixtureB.GetType();

	if (box2d.ENABLE_ASSERTS) { box2d.b2Assert(0 <= typeA && typeB < box2d.b2ShapeType.e_shapeTypeCount); }
	if (box2d.ENABLE_ASSERTS) { box2d.b2Assert(0 <= typeA && typeB < box2d.b2ShapeType.e_shapeTypeCount); }

	var reg = this.m_registers[typeA][typeB];

	var destroyFcn = reg.destroyFcn;
	destroyFcn(contact, this.m_allocator);
}

